#ifndef EIGEN_AVX512_SPECIALFUNCTIONS_H
#define EIGEN_AVX512_SPECIALFUNCTIONS_H

namespace Eigen {
namespace internal {

EIGEN_INSTANTIATE_SPECIAL_FUNCS_F16(Packet16f, Packet16h)
EIGEN_INSTANTIATE_SPECIAL_FUNCS_BF16(Packet16f, Packet16bf)

}  // namespace internal
}  // namespace Eigen

#endif  // EIGEN_AVX512_SPECIALFUNCTIONS_H
